: <<sound.f>> ;

LoadLibrary winmm.dll CONSTANT winmm

$1     CONSTANT WHDR_DONE
$2     CONSTANT WHDR_PREPARED
$3BD   CONSTANT MM_WOM_DONE
$30000 CONSTANT CALLBACK_FUNCTION
$10000 CONSTANT CALLBACK_WINDOW

6 winmm WINAPI: waveOutOpen
3 winmm WINAPI: waveOutUnprepareHeader
3 winmm WINAPI: waveOutPrepareHeader
3 winmm WINAPI: waveOutWrite
1 winmm WINAPI: waveOutClose
1 winmm WINAPI: waveOutPause

-1 CONSTANT WAVE_MAPPER
$AC44 CONSTANT SAMPLINGRATE                                                  \ 44.1 Khz
$10   CONSTANT SAMPLINGDEPTH                                                 \ 16 bit
SAMPLINGDEPTH 8 / CONSTANT SAMPLINGLEN   \ 16bit = 2 byte
$3E8  CONSTANT SAMPLINGDURATION           \ each buffer hold 1000ms, 2buffer = 2sec
SAMPLINGDURATION SAMPLINGRATE * 1000 / 
SAMPLINGLEN * 2/ CONSTANT WAVEBUFSIZE     \ buffer size to hold the 1/2 wave form

8 CELLS CONSTANT WAVEHDRSIZE 
                                           \ use double buffer to avoid *blips*
CREATE 2*WAVEBUF WAVEBUFSIZE 2* ALLOT      \ two wave form buffer
CREATE 2*WAVEHDR WAVEHDRSIZE 2* ALLOT      \ header for two buffer
2*WAVEBUF WAVEBUFSIZE 2* 0 FILL            \ zero fill
2*WAVEHDR WAVEHDRSIZE 2* 0 FILL
VARIABLE hWaveOut
VARIABLE 'WAVEGENERATOR

VARIABLE TOGGLEBUF                         \ buffer selected , 0 = upper half , 1 = lower half

: NEXTBUF
  1 TOGGLEBUF @ - TOGGLEBUF !
; 

: WAVEBUF                                  \ return address of current buffer, selected by TOGGLEBUF
  2*WAVEBUF TOGGLEBUF @ WAVEBUFSIZE * +
;

: WAVEHDR                                  \ return address of current header, selected by TOGGLEBUF
  2*WAVEHDR TOGGLEBUF @ WAVEHDRSIZE * +
;

: WAVEHDR.FLAG                             \ return address of flag  of current header
  WAVEHDR 4 CELLS +
;

: WAVEHDR.DATA                             \ waveform buffer
  WAVEHDR 
;

: WAVEHDR.SIZE                             \ waveform buffer size
  WAVEHDR CELL+
;
                                     \ waveform format defination
CREATE WAVEFORMATEX
       $1 C, 0 C,                    \ PCM format
       $1 C, 0 C,                    \ one channel 
       SAMPLINGRATE ,                \ sampling rate
       SAMPLINGRATE SAMPLINGLEN * ,  \ byte per second
       SAMPLINGLEN C, 0 C,           \ block align
       SAMPLINGDEPTH C, 0 C,         \ 16 bit
       0  C, 0 C,                    \ extra 

: FILLWAVEBUFFER ( -- )
  WAVEHDR.FLAG @ WHDR_PREPARED AND IF     \ Unprepare header if the buffer is already prepared
     hWaveOut @ WAVEHDR WAVEHDRSIZE waveOutUnprepareHeader DROP
  THEN

  \ run user define function to fill the buffer
  WAVEBUF WAVEBUFSIZE 'WAVEGENERATOR @ ?DUP IF EXECUTE ELSE 2DROP THEN

  \ Prepare the buffer to be sent to the WaveOut API
  WAVEBUF     WAVEHDR.DATA !
  WAVEBUFSIZE WAVEHDR.SIZE !

  hWaveOut @ WAVEHDR WAVEHDRSIZE waveOutPrepareHeader DROP

  hWaveOut @ WAVEHDR WAVEHDRSIZE waveOutWrite DROP           \ Send the buffer the the WaveOut queue

  NEXTBUF		                             \ switch to next buffer
;

\ callback for wave out
: waveOutProc  (  hwo uMsg Instance Param1 Param2 -- return )
  2DROP DROP                                \ hwo hMsg 
  MM_WOM_DONE = IF  FILLWAVEBUFFER THEN     \ when wave mapper run out of buffer , fill next buffer
  DROP                                      \ drop hwo
  0                                         \ return value for void
;

5 ' waveOutProc CALLBACK: &waveOutProc

: waveopen
  hWaveOut WAVE_MAPPER WAVEFORMATEX &waveOutProc 1 CALLBACK_FUNCTION
  waveOutOpen IF
    ." Error opening wave device"  EXIT
  THEN

\ fill up two buffer, when one buffer run out, callback will tell us to fill it
\ and 
  FILLWAVEBUFFER
  FILLWAVEBUFFER
;

: waveclose

  hWaveOut @ WAVEHDR WAVEHDRSIZE waveOutUnprepareHeader DROP 
  NEXTBUF
  hWaveOut @ WAVEHDR WAVEHDRSIZE waveOutUnprepareHeader DROP
  NEXTBUF
  
  hWaveOut @ waveOutPause DROP        \ Reset will halt the thread , so use Pause
  hWaveOut @ waveOutClose DROP
;

\ ================================== USER CODE STARTS HERE =================================================

VARIABLE STEPS
\ other thread will enter this routine, 
\ since user variable is stored in thread local storage,
\ the callback is not able to access user variable, 
\ do not use word using user variable
\ data stack item is limited to 64
: MY_WAVE ( BUF BUFSIZE -- )
  $100 / STEPS !                   \ split to 256 segments
  STEPS @ 
  FOR AFT
    DUP STEPS @
    R@ $3 *                         \ wave output level
    FILL  
    $100 +                          \ next segment
  THEN
  NEXT
  DROP
;

\ assign cfa to 
' MY_WAVE 'WAVEGENERATOR !
CR .( 
waveopen  : start making noise
waveclose : stop making noise )